/**
 * \file: mpimain.cpp
 *
 *
 * Simple MPI communication example.
 *
 * The silly policeman is moving randomly and 
 * shouting to his radio: "Burglar" if he sees the robber.
 * The others can hear this, but they don't know where 
 * the robber is because his fool does not say it.
 *
 * Compile: 
 *  make mpimain 
 * 
 * Executing: 
 *  mpitest.sh
 *
 */

#include <iostream>
#include <sstream>
#include <string>
#include <stdlib.h>
#include <typeinfo>
#include <mpi.h>

#include "player.h"

using std::cin;

#define TAG_RADIO 1	// Radio communication

/// microsecond sleep
/// @param t - time to sleep in useconds
//  @return -1 if signal has been received, 0 otherwise
int uSleep(int usec) {
    struct timespec tim = { 0, usec*1000L }, trem;
    return nanosleep(&tim, &trem);
}

/// debMsg class for debugging.
/// Collects the output and inserts to the stream when the flush() is called.
/// In this way the outputs of differnt parallel tasks will not mixed (hopefully). 
class debMsg : public std::stringstream {
    std::ostream& os;
public:
    debMsg(std::ostream& s = std::cerr) :os(s) {}
    void flush() {
	(os << this->str()).flush();
	this->str("");
    }
    ~debMsg() { flush(); }
}; 

/// My endl
class endl_t {} endl;

/// My endline inserted to the stream
std::ostream& operator<<(std::ostream& os, endl_t) {
    os.put('\n');	// newline
    try {		
        dynamic_cast<debMsg&>(os).flush(); // convertable -> calls flush()
    } catch (std::bad_cast) {
        os.flush();	// can't cast 
    }
    return os;
}

int main(int argc, char ** argv) {
    debMsg cerr;	// this is the own cerr, not from std!!!

    try {
        MPI::Init(argc, argv);
        int rank = MPI::COMM_WORLD.Get_rank();
        int size = MPI::COMM_WORLD.Get_size();
    
        if(argc != 3) {
            cerr << "Usage: " << argv[0] << " <hostname> <port>" << endl;
            MPI::Finalize();
            return 1;;
        }
    
        // processing the command line arguments
        std::stringstream ss(argv[2]);
        int port;
        ss >> port; // first argument is the port number
    
        // starting the policeman
        Player pl(argv[1], port, Player::cop);
        cerr << "Initial position of cop #" << rank 
             << " (" << pl.getx() << "," << pl.gety() << ")" << endl;
    
        while(1) {
            uSleep(100);        // just for slowing down the process
    
            MPI::Status stat;
            if (MPI::COMM_WORLD.Iprobe(MPI::ANY_SOURCE, TAG_RADIO, stat)) { // message is ready
		int len = stat.Get_count(MPI::CHAR);
                char buff[100];
                MPI::COMM_WORLD.Recv(buff, len, MPI::CHAR, stat.Get_source(), TAG_RADIO);
                cerr << "Cop #" << rank << " got a message from cop #" 
                     << stat.Get_source() << ": " << buff << endl;
            }
    
            int r = rand() % (Player::right + 1); // random movement
            Player::direction d;
            if (r == 0) d = Player::up;
            else if (r == 1) d = Player::down;
            else if (r == 2) d = Player::left;
            else d = Player::right;

            char *buffer = pl.move(d);	// move command

	    // now we check the answer
            for (char *p = buffer; *p; p++) {
                if (*p == '1') {       // It is a robber
                   // shouting to the radio:
	           char msg[] = "Burgar";
                   for (int i = 0; i < size; i++) {  // to all
                        if (i != rank)
                            MPI::COMM_WORLD.Send(msg, sizeof(msg), MPI::CHAR, i, TAG_RADIO);
                   }
                   break;                 // we do not expect mot than one robber on the street
	  	}
            }
            delete[] buffer;
        }
        MPI::Finalize();
    } catch (std::runtime_error &err) {
	cerr << err.what() << endl;
	MPI::COMM_WORLD.Abort(1);
    } catch (MPI::Exception &err) {	// this generally not enabled
	cerr << err.Get_error_string() << endl;
        MPI::COMM_WORLD.Abort(1);
    }
    return 0;
}
